#########################################################
## Prolific Technology Inc.
##
## Project: PL-1029 NOR Flash Boot Code
##
## Author: Jun-Chen
##
## Current Version: 1.0.2
##
## History:
##    2006-01-05: initial		Ver. 1.0.2
##
##
#########################################################

.include "./include/macro_sys.h"
.include "./include/macro_memory.h"
.include "./include/reg_info.h"
#include "./include/config.h"

/*-----------------------------------------------------------
 System error code
-----------------------------------------------------------*/
.equ CPU_FAULT,				0xE0	@CPU hardware error
.equ CPU_HW_Exception,			0xE1	@CPU hardware exception
.equ ERR_MCAS,				0xE2	@Memory configuration error
.equ ERR_MRAS,				0xE3	@Memory configuration error
.equ ERR_READ_SM,			0xE4	@On-board flash read error
.equ ERR_UNBOOTABLE_SM,			0xE5	@On-board flash unbootable
.equ BAUDRATE_TOO_HIGH,			0xE6	@Console access error
.equ ERR_CONSOLE_DETECT,		0xE7	@Console access error
.equ ERR_LESS_MEM,			0xE8	@Memory configuration error
.equ ERR_MEM_TEST,			0xE9	@Memory test error
.equ ERR_SERIAL_BREAK,			0xEA	@Console access error
.equ ERR_FRAME,				0xEB	@Console access error
.equ ERR_PARITY,			0xEC	@Console access error
.equ ERR_OVERRUN,			0xED	@Console access error
.equ ERR_READ_SEEPROM,			0xEE	@Serial EEPROM read error
.equ ERR_UNBOOTABLE_SEEPROM,          	0xEF	@Serial EEPROM read error
.equ ERR_READ_NOR,			0xF1	@NOR flash read error
.equ ERR_UNBOOTABLE_NOR,		0xF2	@NOR flash unbootable

.global _start
.global console_flow

.global SM_Init
.global SM_Read_Page
.global SM_Get_Phy_Addr
.global Show_Status
.global Show_Message
.global Show_Register

.global console_flow
.global Version
.global Nand_Flash_Boot

@ ------------------------------------------------------------------------


.code 32
.text
_start:


	b	close_INT
	ldr	r0, =_nor_offset
@ ------------------------------------
@ ARM processor exception vector table
@ ------------------------------------
.LCvectors:
ev00:	ldr pc,[pc,#0x18]		@ Reset exception
ev04:	ldr pc,[pc,#0x18]		@ Undefined Instruction exception
ev08:	ldr pc,[pc,#0x18]		@ Software Interrupt exception
ev0C:	ldr pc,[pc,#0x18]		@ Prefetch Abort exception
ev10:	ldr pc,[pc,#0x18]		@ Data Abort exception
ev14:	ldr pc,[pc,#0x18]		@ (Not used)
ev18:	ldr pc,[pc,#0x18]		@ Interrupt exception
ev1C:	ldr pc,[pc,#0x18]		@ Fast Interrupt exception

@----------------------------------------------------------------
@ EXCEPTION HANDLE
@----------------------------------------------------------------
vector_reset:
	m_show_status 0xD0
	ldr pc,=ROM_BASE_ADDRESS
vector_undefined:
	m_show_status 0xD1
	b vector_undefined
vector_swi:
	m_show_status 0xD2
	b vector_swi
vector_prefetch:
	m_show_status 0xD3
	b vector_prefetch
vector_data:
	m_show_status 0xD4
	b vector_data
vector_addrerror:
	m_show_status 0xD5
	b vector_addrerror
vector_irq:
	m_show_status 0xD6
	b vector_irq
vector_fiq:
	m_show_status 0xD7
	b vector_fiq
@----------------------------------------------------------------

.align	2


close_INT:
	m_disable_CPU_interrupt		@ macro_sys.h
	m_show_status 0x35
  	m_disable_INTC_interrupt	@ macro_sys.h
	m_show_status 0x36

clock_setting:
	m_set_clock			@ macro_sys.h
	m_show_status 0x37

memory_sizing:
	m_memory_sizing			@ macro_memory.h
	m_show_status 0x38

set_exception_entry:
	.equ	__ref_addr, ev1C-ev00+4
	ldr     lr,=0x00000000
	ldr	r1,=.LCvectors		@ write ev00..ev1c to address 0x00000000
	ldmia	r1, {r0, r1, r2, r3, r4, r5, r6, r7}
	stmia	lr, {r0, r1, r2, r3, r4, r5, r6, r7}

	ldr	r0,=__ref_addr		@ write ev00 reference address
	ldr 	r1,=vector_reset
	str 	r1,[r0],#4
	ldr 	r1,=vector_undefined	@ write ev04 reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_swi		@ write ev08 reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_prefetch	@ write ev0c reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_data		@ write ev10 reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_addrerror	@ write ev14 reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_irq		@ write ev18 reference address
	str 	r1,[r0],#4
	ldr 	r1,=vector_fiq		@ write ev1c reference address
	str 	r1,[r0]
	m_show_status 0x39

SP_init:
	m_setting_SP STACK_ADDR		@ macro_sys.h
	m_show_status 0x3a


nor2ram:
	@  get total size (r6) 
	ldr	r0,=_fproboot
	ldr	r1,=_edata
	sub	r6,r1,r0
	@  get nor start address (r1)
	ldr	r1,=_ebootsect
	@  get mem start address (r0)
	ldr	r0,=_fproboot
	
	@  start copy
	mov	r2,#0
cp_loop:
	ldr	r3,[r1],#4
	str	r3,[r0],#4
	add	r2,r2,#4
	cmp	r2,r6
	blt	cp_loop
		
/*
	@
	@  get total size (r6) 
	@
	ldr	r0,=_fproboot
	ldr	r1,=_edata
	sub	r6,r1,r0
	bic	r6,r6,#0xf
	add	r6,r6,#0x10
	@
	@  nor (ppi-amd) init 
	@
	ldr	r0,=SOFT_RST_REG	@ software_reset, writel(0x80000000,SOFT_RST_REG);
	mov	r1,#(1<<31)
	str	r1,[r0]
	ldr	r0,=CONFIG_REG		@ Open RB exist, writel((readl(CONFIG_REG)|0x10),CONFIG_REG);
	ldr	r1,[r0]
	orr	r1,r1,#(1<<4)
	str	r1,[r0]
	ldr	r0,=PIO_DATA_REG	@ amd_reset, writel(0xF0,PIO_DATA_REG);
	mov	r1,#0xF0
	str	r1,[r0]
	ldr	r0,=INTR_M_REG		@ interrupt_mask(0x00), writeb(mask,INTR_M_REG);
	mov	r1,#0
	strb	r1,[r0]
	ldr	r0,=CONFIG_REG		@ r7=update_byte_enable();
	ldr	r1,[r0]
	mov	r1,r1,LSR #30
	and	r7,r1,#1	
	@
	@  setup DMAC CH-14 ; setup_dma(DMA_PL_NORFLASH, UP_STREAM, dma_up_bufp, icopy);
	@	
	mov	r0,#0x400		@ set MDT address 0x400 
	ldr	r1,=_fproboot
	str	r1,[r0],#4		@ MDT[0]=mem start address
	orr	r1,r6,#0xC0000000
	str	r1,[r0]			@ MDT[1]=0xC000_0000 | SIZE
	mov	r0,#0x400		@ writel(0x400,DMA_REG_MDT+Ch*4);
	ldr	r1,=0x197c0078
	str	r0,[r1]
	mov	r0,#0xC8000000		@ writel(0xC800_0000,DMA_REG_CHN+Ch*4);
	ldr	r1,=0x197c0038
	str	r0,[r1]
	@
	@  setup NOR Flash DMA	 
	@	
	ldr	r0,=NOR_ADDR_REG	@ writel(addr,NOR_ADDR_REG);//nor start address
	ldr	r1,=_fbootsect
	ldr	r2,=_ebootsect
	sub	r1,r2,r1
	str	r1,[r0]
	ldr	r0,=DMA_S_REG		@ writel((0x40000000|icopy),DMA_S_REG);
	orr	r1,r6,#0x40000000
	str	r1,[r0]
	@
	@  polling NOR Flash Status	 
	@	
	ldr	r0,=CTRL_ST_REG		@ wait_INTR(INT_DMA,0)
pol_st:	ldrb	r1,[r0]
	ands	r1,r1,#1	
	bne	pol_st	
*/	
	
norboot:
/*
#if CONFIG_CACHE_ON
        bl      __cache_on
#endif
        m_show_status 0x3b
	b	init
*/
	m_show_status 0x3b
	ldr pc,=_fproboot
	.end




