/***************************************************************
*             Copyright (C) 2003 by SerComm Corp.
*                    All Rights Reserved.
*
*      Use of this software is restricted to the terms and
*      conditions of SerComm's software license agreement.
*
*                        www.sercomm.com
****************************************************************/
#ifndef _DDNS_H_
#define _DDNS_H_

#include "sc_config.h"

#define DDNS_MAX_HOSTNAME_LEN	128
#define GENERIC_STRING_LEN	32
#define DDNS_ACCOUNT_LEN	64	//48, accont in tzo is email address
#define DDNS_PASSWORD_LEN	32
#define DDNS_OK		0
#define DDNS_ERROR	-1
#define DDNS_EINVALID   -2	/* Malformat Input */


#define ENABLE		1
#define DISABLE		0

#define DDNS_UNIT_MIN 	1
#define DDNS_UNIT_HR	2
#define DDNS_UNIT_DAY	3

#ifdef _SC_TWO_DDNS_ALL_EXIST_
#define DDNS_SOLOLINK_SERVER	0
#endif

/****************************************************************
 * The usage of schedule can be viewed as the the unit of time, 
 * and the interval is the value of that unit. For example, 
 * schedule=DDNS_UNIT_MIN,interval=20 -> update every 20 minutes
 * schedule=DDNS_UNIT_HOUR,interval=6 -> update every 6 hours
 *****************************************************************
 * The start_hour and the start_min is used to indicate the 
 * starting time, the range of hour is from 0 to 23 , and the range
 * of minute is from 0 to 59
 *****************************************************************/
 
typedef struct ddns_param
{
  int enable;				/* flag to indicate en/dis-able  */
  int server_id;	/* service providing server name */
  					/* server must be		 * 
					 * DYNDNS_NAME or TZO_NAME	 */
  char account[DDNS_ACCOUNT_LEN+1];	/* username / account 		 */
  char password[DDNS_PASSWORD_LEN+1];	/* passowrd for authentication	 */
  char hostname[DDNS_MAX_HOSTNAME_LEN+1]; /* domain name		 */
  int schedule; 		 	/* scheduling unit		 */
  int interval;				/* scheduling interval		 */
  int start_hour;			/* starting hour 0 - 23 	 */
  int start_min;			/* starting minutes 0 - 59	 */
  char outip[GENERIC_STRING_LEN+1];	/* outer ip address 		 */
  char status[DDNS_ACCOUNT_LEN+1];	/* check by logfile 		 */
}ddns_param_t;

/***********************************************************
 * routine: read_ddns_config
 * description: read configuration file from system.conf
 * input: 	config	
 * output:	config
 * return value: DDNS_OK
 *		 DDNS_ERROR
 ***********************************************************/ 	
int read_ddns_config(ddns_param_t *config);

/***********************************************************
 * routine: write_ddns_config
 * description: write configuration file to system.conf 
 *		and del crontab entry / add set_crontab 
 * input: 	config	
 * output:	config
 * return value: DDNS_OK
 *		 DDNS_ERROR
 ***********************************************************/ 	
int write_ddns_config(ddns_param_t *config);

#endif
