/***************************************************************
* (C) COPYRIGHT SerComm Corporation All Rights Reserved, 2004.
*
*      Use of this software is restricted to the terms and
*      conditions of SerComm's software license agreement.
*
*                        www.sercomm.com
****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdint.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h> 
#include <unistd.h>
#include <linux/rtc.h>
#include "sc_config.h"
#include "time.h"

#include "sc_debug.h"

#ifdef _S35390A_RTC_	// Formal RTC driver in kmods

int set_RTC_time(unsigned char *time)
{
	struct rtc_time rtime;
	int fd = -1;

    memset(&rtime,0,sizeof(struct rtc_time));
    rtime.tm_year=(time[4]*10+time[5])%100;   // Year
    rtime.tm_mon=time[6]*10+time[7] +1;   // Month	
    rtime.tm_mday=time[8]*10+time[9];   // Day	
    rtime.tm_wday=0x00; 		       // Week
    rtime.tm_hour=(time[10]*10+time[11])&0x3F; // Hour
    rtime.tm_min=time[12]*10+time[13]; // Minute
    rtime.tm_sec=time[14]*10+time[15]; // Second

/*
	SC_DBG_CGI("%s(%d): The date %x%x%x%x / %x%x / %x%x , week %x%x, and time %x%x:%x%x:%x%x\n", 
			__FUNCTION__, __LINE__, 
			time[0],time[1],time[4],time[5],
			time[6],time[7],
			time[8],time[9],
			time[2],time[3],
			time[10],time[11],
			time[12],time[13],
			time[14],time[15]);
*/	
    fd = open("/dev/rtc_drv", O_RDWR|O_EXCL);
/*
	SC_DBG_CGI("%s(%d): fd=%d, In RTC: [%d,%d,%d,%d,%d,%d,%d]\n", 
				__FUNCTION__, __LINE__, fd, 
				rtime.tm_year,
				rtime.tm_mon,
				rtime.tm_mday,
				rtime.tm_wday,
				rtime.tm_hour,
				rtime.tm_min,
				rtime.tm_sec);
*/				
	if (fd > 0)
	{
    	ioctl(fd, RTC_SET_TIME,&rtime); 
    	close(fd);
		return 0;
	}
	else
	{
		return -1;
	}
}
void read_RTC_time(unsigned char *time_rtc)
{
	struct rtc_time rtime;
	int fd = -1;

	//SC_DBG_CGI("%s(%d)\n", __FUNCTION__, __LINE__);
    memset(&rtime,0,sizeof(struct rtc_time));
    fd = open("/dev/rtc_drv", O_RDONLY|O_EXCL);
	if (fd > 0)
	{
	    ioctl(fd, RTC_RD_TIME,&rtime); 
	    close(fd);

/*
	SC_DBG_CGI("%s(%d): fd=%d, In RTC: [%d,%d,%d,%d,%d,%d,%d]\n", 
				__FUNCTION__, __LINE__, fd, 
				rtime.tm_year,
				rtime.tm_mon,
				rtime.tm_mday,
				rtime.tm_wday,
				rtime.tm_hour,
				rtime.tm_min,
				rtime.tm_sec);
*/		
	    rtime.tm_year += 100;
	    time_rtc[4]=rtime.tm_year/10;
	    time_rtc[5]=rtime.tm_year%10;
	    time_rtc[6]=rtime.tm_mon/10;
	    time_rtc[7]=rtime.tm_mon%10;
	    time_rtc[8]=rtime.tm_mday/10;
	    time_rtc[9]=rtime.tm_mday%10;
	    time_rtc[10]=(rtime.tm_hour&0x3F)/10;
	    time_rtc[11]=(rtime.tm_hour&0x3F)%10;
	    time_rtc[12]=rtime.tm_min/10;
	    time_rtc[13]=rtime.tm_min%10;
	    time_rtc[14]=rtime.tm_sec/10;
	    time_rtc[15]=rtime.tm_sec%10;

/*
	SC_DBG_CGI("%s(%d): The date %x%x%x%x / %x%x / %x%x , week %x%x, and time %x%x:%x%x:%x%x\n", 
			__FUNCTION__, __LINE__, 
			time_rtc[0],time_rtc[1],time_rtc[4],time_rtc[5],
			time_rtc[6],time_rtc[7],
			time_rtc[8],time_rtc[9],
			time_rtc[2],time_rtc[3],
			time_rtc[10],time_rtc[11],
			time_rtc[12],time_rtc[13],
			time_rtc[14],time_rtc[15]);
*/			
	}
	return;
}

#else

#define I2C_DEVICE	"/dev/i2c/0"
#define SLAVE_ADDR_READ 0x6F

//#define RTCDEBUG	0
#ifdef RTCDEBUG
#define rtc_debug(format, argument...)	printf(format, ## argument);	
#else
#define rtc_debug(format, argument...)	
#endif

#define BCD2BIN(val)	(((val) & 0x0f) + ((val)>>4)*10)
#define BIN2BCD(val)	((((val)/10)<<4) + (val)%10)

/* offsets into CCR area */

#define CCR_SEC			0
#define CCR_MIN			1
#define CCR_HOUR		2
#define CCR_MDAY		3
#define CCR_MONTH		4
#define CCR_YEAR		5
#define CCR_WDAY		6
#define CCR_Y2K			7

#ifdef  _ISL1208_RTC_
#define X1205_REG_SR		0x07	/* status register */
#define X1205_REG_DW		0x06
#define X1205_REG_YR		0x05
#define X1205_REG_MO		0x04
#define X1205_REG_DT		0x03
#define X1205_REG_HR		0x02
#define X1205_REG_MN		0x01
#define X1205_REG_SC		0x00
#define X1205_CCR_BASE		0x00	/* Base address of CCR */
#define X1205_SR_WEL		0x10	/* Write Enable Latch */
#define X1205_DUMY_WORD	0x00
#define X1205_HR_MIL		0x80	/* Set in ccr.hour for 24 hr mode */
#define REG_NUM_WR              7
#else
#define X1205_REG_SR		0x3F	/* status register */
#define X1205_REG_Y2K		0x37
#define X1205_REG_DW		0x36
#define X1205_REG_YR		0x35
#define X1205_REG_MO		0x34
#define X1205_REG_DT		0x33
#define X1205_REG_HR		0x32
#define X1205_REG_MN		0x31
#define X1205_REG_SC		0x30
#define X1205_CCR_BASE		0x30	/* Base address of CCR */
#define X1205_SR_WEL		0x02	/* Write Enable Latch */
#define X1205_SR_RWEL		0x04	/* Register Write Enable */
#define X1205_DUMY_WORD		0x00
#define X1205_HR_MIL		0x80	/* Set in ccr.hour for 24 hr mode */
#define REG_NUM_WR              8
#endif

#ifndef _SC_NO_RTC_
#ifdef _X1206_SET_RTC_TIME_TWICE_
void set_RTC_time_1(unsigned char *time);
void set_RTC_time(unsigned char *time)
{
	set_RTC_time_1(time);
	set_RTC_time_1(time);
}

void set_RTC_time_1(unsigned char *time)
#else
void set_RTC_time(unsigned char *time)
#endif
{
	int fd=-1;
    
	struct rtc_time pl1029_rtime;
#ifdef _SC_SUPPORT_2_RTC_AUTO_
    char rtime[8];
	int ret=-1,i;
#ifdef _ISL1208_RTC_
	static const unsigned char wel[2] = {X1205_REG_SR, X1205_SR_WEL };
	static const unsigned char diswe[2] = {X1205_REG_SR, 0 }; 
#else
	static const unsigned char wel[3] = { 0, X1205_REG_SR, X1205_SR_WEL };

	static const unsigned char rwel[3] = { 0, X1205_REG_SR, X1205_SR_WEL | X1205_SR_RWEL };
	static const unsigned char diswe[3] = { 0, X1205_REG_SR, 0 }; 
#endif

	rtc_debug("%s(%d)\n", __FUNCTION__, __LINE__);

	fd = open(I2C_DEVICE,O_RDWR);
	if (fd < 0)
		goto x1205_set_fail;

	if (ioctl(fd,0x0706,SLAVE_ADDR_READ) < 0) 
	{
		rtc_debug("Slave address %x\n",SLAVE_ADDR_READ);
		goto x1205_set_fail;
	}

	ret=write(fd,wel,sizeof(wel));
	if(ret != sizeof(wel)) 
	{
		rtc_debug("Write enable latch:%d\n",ret);
		goto x1205_set_fail;
	} 

#ifndef _ISL1208_RTC_
	ret=write(fd,rwel,3);
	if(ret != 3) 
	{
		rtc_debug("Register write enable latch:%d\n",ret);
		goto x1205_set_fail;
	}
#endif

	rtime[CCR_Y2K]=BIN2BCD(time[0]*10+time[1]);	   // Y2K
	rtime[CCR_YEAR]=BIN2BCD((time[4]*10+time[5])%100);   // Year
	rtime[CCR_MONTH]=BIN2BCD((time[6]*10+time[7])+1);  // RTC month 1-12	 // because application use month 0-11, but RTC use 1-12, add one month before set to RTC
	rtime[CCR_MDAY]=BIN2BCD(time[8]*10+time[9]);   // Day	
	rtime[CCR_WDAY]=BIN2BCD(0x00); 		  // Week
	rtime[CCR_HOUR]=BIN2BCD(time[10]*10+time[11]) | X1205_HR_MIL; // Hour
	rtime[CCR_MIN]=BIN2BCD(time[12]*10+time[13]); // Minute
	rtime[CCR_SEC]=BIN2BCD(time[14]*10+time[15]); // Second

	for(i = 0; i < REG_NUM_WR; i++)
	{
#ifdef _ISL1208_RTC_
		unsigned char rdata[2] = {X1205_CCR_BASE + i, rtime[i] };
		ret=write(fd,rdata,2);
		rtc_debug("Reg:%x - Value:%x\n",rdata[0],rdata[1]);		
#else
		unsigned char rdata[3] = { 0, X1205_CCR_BASE + i, rtime[i] };
		ret=write(fd,rdata,3);
		rtc_debug("Reg:%x - Value:%x\n",rdata[1],rdata[2]);		
#endif
		if(ret != sizeof(rdata))
			goto x1205_set_fail;
	}

	ret=write(fd,diswe,sizeof(diswe));
	if(ret != sizeof(diswe))
	{
		rtc_debug("Disable write latch:%d\n",ret);
		goto x1205_set_fail;
	}

	close(fd);

	rtc_debug("%s(%d): Use the external RTC\n", __FUNCTION__, __LINE__);
	{
	    return;
	}

x1205_set_fail:
	if (fd > 0)
	{
    	close(fd);
	}
    rtc_debug("Fail to set date to X1205, and use the internal RTC\n");
#endif
   // set X1205 failure,use PL1029 rtc
    memset(&pl1029_rtime,0,sizeof(struct rtc_time));
    pl1029_rtime.tm_year=time[4]*10+time[5];   // Year
    pl1029_rtime.tm_mon=time[6]*10+time[7] +1;   // Month	
    pl1029_rtime.tm_mday=time[8]*10+time[9];   // Day	
    pl1029_rtime.tm_wday=0x00; 		       // Week
    pl1029_rtime.tm_hour=time[10]*10+time[11]; // Hour
    pl1029_rtime.tm_min=time[12]*10+time[13]; // Minute
    pl1029_rtime.tm_sec=time[14]*10+time[15]; // Second

    fd = open("/dev/rtc", O_RDWR|O_EXCL);

    ioctl(fd, RTC_SET_TIME,&pl1029_rtime); 

    close(fd);

   return  ;
}

void read_RTC_time(unsigned char *time_rtc)
{
	struct rtc_time pl1029_rtime;
	int fd=-1;
#ifdef _SC_SUPPORT_2_RTC_AUTO_
	int ret=-1,value;
	char rtime[8];
#ifdef _ISL1208_RTC_
	char sram = X1205_CCR_BASE;
#else
	char sram[2]={X1205_DUMY_WORD,X1205_CCR_BASE};
#endif
	rtc_debug("%s(%d)\n", __FUNCTION__, __LINE__);

	fd = open(I2C_DEVICE,O_RDWR);
	if (fd < 0)
		goto x1205_get_fail;

	if (ioctl(fd,0x0706,SLAVE_ADDR_READ) < 0) 
	{
		rtc_debug("Slave address %x\n",SLAVE_ADDR_READ);
		goto x1205_get_fail;
	}
	
#ifdef _ISL1208_RTC_
	ret = write(fd, &sram, 1);
	if(ret != 1)
	{
		rtc_debug("write addr %d:%d\n",sram, ret);
		goto x1205_get_fail;
	}
#else
	ret = write(fd, sram, 2);
	if(ret != 2)
	{
		rtc_debug("write addr %d:%d\n",sram[1],ret);
		goto x1205_get_fail;
	}
#endif

	ret= read(fd, rtime, REG_NUM_WR);
	if(ret != REG_NUM_WR)
	{	
#ifdef _ISL1208_RTC_
		rtc_debug("read addr %d:%d\n",sram,ret);
#else
		rtc_debug("read addr %d:%d\n",sram[1],ret);
#endif
		goto x1205_get_fail;
	}	

    rtc_debug("raw read data - sec=%02d, min=%02d, hr=%02d, "
		"mday=%02d, mon=%02d, year=%02d, wday=%02d, y2k=%02d\n",
		BCD2BIN(rtime[CCR_SEC]) , BCD2BIN(rtime[CCR_MIN]), BCD2BIN(rtime[CCR_HOUR]&0x3F), BCD2BIN(rtime[CCR_MDAY]),BCD2BIN(rtime[CCR_MONTH]), BCD2BIN(rtime[CCR_YEAR]), BCD2BIN(rtime[CCR_WDAY]), BCD2BIN(rtime[CCR_Y2K]));

#if 0
    time_rtc[0]=BCD2BIN(rtime[CCR_Y2K])/10;
    time_rtc[1]=BCD2BIN(rtime[CCR_Y2K])%10;
    time_rtc[4]=BCD2BIN(rtime[CCR_YEAR])/10;
    time_rtc[5]=BCD2BIN(rtime[CCR_YEAR])%10;
    time_rtc[6]=BCD2BIN(rtime[CCR_MONTH])/10;	/* Mon is 1-12 */
    time_rtc[7]=BCD2BIN(rtime[CCR_MONTH])%10;
#endif
    value = BCD2BIN(rtime[CCR_YEAR]) + 100;
    time_rtc[4]=value/10;
    time_rtc[5]=value%10;
    value = BCD2BIN(rtime[CCR_MONTH]);
    //value += 1;
    time_rtc[6]=value/10;
    time_rtc[7]=value%10;
    time_rtc[8]=BCD2BIN(rtime[CCR_MDAY])/10;
    time_rtc[9]=BCD2BIN(rtime[CCR_MDAY])%10;
    time_rtc[10]=BCD2BIN(rtime[CCR_HOUR]&0x3f)/10;
    time_rtc[11]=BCD2BIN(rtime[CCR_HOUR]&0x3f)%10;
    time_rtc[12]=BCD2BIN(rtime[CCR_MIN])/10;
    time_rtc[13]=BCD2BIN(rtime[CCR_MIN])%10;
    time_rtc[14]=BCD2BIN(rtime[CCR_SEC])/10;
    time_rtc[15]=BCD2BIN(rtime[CCR_SEC])%10;

    close(fd);

	rtc_debug("%s(%d): Use the external RTC\n", __FUNCTION__, __LINE__);
    return;

x1205_get_fail:
	if (fd > 0)
	{
   		close(fd);
	}
   rtc_debug("Fail to get data from X1205, and use the internal RTC\n");
#endif
   // read X1205 failure,use PL1029 rtc
    memset(&pl1029_rtime,0,sizeof(struct rtc_time));
 	
    fd = open("/dev/rtc", O_RDONLY|O_EXCL);
	
    ioctl(fd, RTC_RD_TIME,&pl1029_rtime); 
    close(fd);
    //pl1029_rtime.tm_mon += 1;
    time_rtc[4]=pl1029_rtime.tm_year/10;
    time_rtc[5]=pl1029_rtime.tm_year%10;
    time_rtc[6]=pl1029_rtime.tm_mon/10;
    time_rtc[7]=pl1029_rtime.tm_mon%10;
    time_rtc[8]=pl1029_rtime.tm_mday/10;
    time_rtc[9]=pl1029_rtime.tm_mday%10;
    time_rtc[10]=pl1029_rtime.tm_hour/10;
    time_rtc[11]=pl1029_rtime.tm_hour%10;
    time_rtc[12]=pl1029_rtime.tm_min/10;
    time_rtc[13]=pl1029_rtime.tm_min%10;
    time_rtc[14]=pl1029_rtime.tm_sec/10;
    time_rtc[15]=pl1029_rtime.tm_sec%10;

   return  ;
}
#else
void set_RTC_time(unsigned char *time)
{
	return;
}
void read_RTC_time(unsigned char *time_rtc)
{
    time_t now;
	struct tm *now_tm;	

	time(&now);
	now_tm=localtime(&now);
	time_rtc[0]=2; /* year */
	time_rtc[1]=0;
	time_rtc[2]=0; /* week */
	time_rtc[3]=5;
	time_rtc[4]=(unsigned char)now_tm->tm_year/10;
	time_rtc[5]=(unsigned char)now_tm->tm_year%10;
	time_rtc[6]=(unsigned char)(now_tm->tm_mon+1)/10;
	time_rtc[7]=(unsigned char)(now_tm->tm_mon+1)%10;
	time_rtc[8]=(unsigned char)now_tm->tm_mday/10;
	time_rtc[9]=(unsigned char)now_tm->tm_mday%10;
	time_rtc[10]=(unsigned char)now_tm->tm_hour/10;
	time_rtc[11]=(unsigned char)now_tm->tm_hour%10;
	time_rtc[12]=(unsigned char)now_tm->tm_min/10;
	time_rtc[13]=(unsigned char)now_tm->tm_min%10;
	time_rtc[14]=(unsigned char)now_tm->tm_sec/10;
	time_rtc[15]=(unsigned char)now_tm->tm_sec%10;

	return ;
	
}
#endif

#if 0
int convert(int range,unsigned char *hi,unsigned char *low){
	
	rtc_debug("range %d\n",range);
	*hi=range/10;
	*low=range%10;

	return 0;
}

int main(int argc,char *argv[]) {
	unsigned char rtc[16],hi_byte,low_byte;

	if(argc==1){
		rtc_debug("rtc set yy mm dd hh mm ss\n");	
		rtc_debug("	y2k:(X-1900)/100 \n");        
		rtc_debug("	yy:0-99 mm:1-12 dd:1-31\n");
		rtc_debug("	hh:0-23 mm:0-59 ss:0-59\n");	
		rtc_debug("rtc read\n");	
		rtc_debug("rtc rw\n");
		exit(0);
	}
	
	if((strcmp(argv[1],"set")==0) && (argc=9))
	{
		convert(atoi(argv[2]),&hi_byte,&low_byte);
		rtc[0]=2; /* Y2K */
		rtc[1]=0;
		rtc[2]=0; /* week */
		rtc[3]=0;

		convert(atoi(argv[3]),&hi_byte,&low_byte);
		rtc[4]=hi_byte;
		rtc[5]=low_byte;

		convert(atoi(argv[4]),&hi_byte,&low_byte);
		rtc[6]=hi_byte;
		rtc[7]=low_byte;

		convert(atoi(argv[5]),&hi_byte,&low_byte);
		rtc[8]=hi_byte;
		rtc[9]=low_byte;

		convert(atoi(argv[6]),&hi_byte,&low_byte);
		rtc[10]=hi_byte;
		rtc[11]=low_byte;

		convert(atoi(argv[7]),&hi_byte,&low_byte);
		rtc[12]=hi_byte;
		rtc[13]=low_byte;

		convert(atoi(argv[8]),&hi_byte,&low_byte);
		rtc[14]=hi_byte;
		rtc[15]=low_byte;

		set_RTC_time(rtc);
	} 
	else if(strcmp(argv[1],"read")==0)
	{
		read_RTC_time(rtc);

	}
	else if(strcmp(argv[1],"rw")==0)
	{
		read_RTC_time(rtc);
		set_RTC_time(rtc);
	}
	else
		rtc_debug("No such command\n");

	return 0;
}	
#endif

#endif

