/***************************************************************
* (C) COPYRIGHT SerComm Corporation All Rights Reserved, 2004.
*
*      Use of this software is restricted to the terms and
*      conditions of SerComm's software license agreement.
*
*                        www.sercomm.com
****************************************************************/
#ifndef  _SYSLOG_MSG_
#define  _SYSLOG_MSG_

#include "sc_config.h"

#define L_OK     0
#define L_ERROR -1

#define SYSLOG_FILE		"/var/nc.log"
#define SYSLOG_IM_DAEMON	"/var/jabber.log"
#if 0
#define SYSLOG_SMTP_DAEMON	"/var/smtp.log"
#define SYSLOG_SNMP_DAEMON	"/var/snmp.log"
#endif
#define SYSLOG_SYSLOG_DAEMON	"/var/syslog.log"

#define SYSLOG_ERROR    0  /* Error log */
#define SYSLOG_WARNING  1  /* Warning log */
#define SYSLOG_NORMAL   2  /* Message log */
#define SYSLOG_ALL	3  /* Internal debug */
#define SYSLOG_FILEONLY	0x1000	/* log file only*/
#define LOG_SERVER_LEN	64
#define IM_SERVER_LEN	64
#define IM_MESSAGE_LEN 128
#define IM_ACCOUT_LEN	64
#define IM_PASSWORD_LEN 64
#define IM_SENDTO_LEN	64

typedef struct SLogConfig {
	int enable;		// 0:disable 1:enable
	char server[LOG_SERVER_LEN+1];
	int port;
} SLogConfig;

typedef struct SYSLOG_INFO
{
	int log_enable;
	SLogConfig log_server;
} SYSLOG_INFO;

#define SYSLOG_SECTION_SMTP	"smtp"
#define SYSLOG_SECTION_DDNS	"DDNS"
#define SYSLOG_SECTION_TRIG	"trigger"
#define SYSLOG_SECTION_SYSTEM                 "system"
#define SYSLOG_SECTION_EZ_IPUPDATE	"ez-ipupdate"
#define SYSLOG_SECTION_HTTP 	"http"
#ifdef _SC_HTTPS_SUPPORT_
#define SYSLOG_SECTION_HTTPS 	"https"
#endif
#define SYSLOG_SECTION_DHCP	"DHCP"
#define SYSLOG_SECTION_NETWORK 	"network"
#define SYSLOG_SECTION_DSP_SRV	"dsp_srv"
#define SYSLOG_SECTION_FTP			"ftp"
#define SYSLOG_SECTION_NTP			"NTP"
#define SYSLOG_SECTION_UPLOAD		"upload"
#define SYSLOG_SECTION_UPNPIGDCP		"UPnPIGDCP"
#ifdef _DDNS_SOLOLINK_
#define SYSLOG_SECTION_IPUPDATE		"ipupdate"
#define SYSLOG_SECTION_LCS_DDNS		"LCS-DDNS"
#endif
#define SYSLOG_SECTION_JABBER		"Jabber"

struct LogJabberConf
{
	int enable;
	char server[IM_SERVER_LEN+1];
	char account[IM_ACCOUT_LEN+1];
	char password[IM_PASSWORD_LEN+1];
	char sendto[IM_SENDTO_LEN+1];
	char message[IM_MESSAGE_LEN+1];
};
struct LogTypeConf
{
	int systemlog_mode;
	int imlog_mode;
	int ftplog_mode;
	int smtplog_mode;
};

#define  HTTP_URL_LEN	       256	
#define  PROXY_SRV_LEN	       	64 
typedef struct HTTPNotifyConfig {
	int  enable;
	char h_url[HTTP_URL_LEN+1];
	char h_proxy_srvname[PROXY_SRV_LEN+1];   
	int  h_proxy_no;			 
	int  method;				 /* 0: Post ,1:Get ?? */
} HTTPNotifyConfig;

int AddLogDbg(const char *format, ...);

int  AddLog(char *task,int level,int index,...); 
int  ClearLog(void);
int SysLogReadData(SYSLOG_INFO *log_info);
int SysLogWriteData(SYSLOG_INFO *log_info);

int LOGReadJabberConf(struct LogJabberConf* conf);
int LOGWriteJabberConf(struct LogJabberConf* conf);
int LOGActionJabberConf(struct LogJabberConf* conf);
//int HTTPNotifyActionConf(void);
int LOGSendMessageJabberConf(char* buf);
int LOGReadLogTypeConf(struct LogTypeConf* conf);
int LOGWriteLogTypeConf(struct LogTypeConf* conf);

/*------------------------------------------------------------------*/
/* purpose: Read HTTP IP address notification from configuration    */ 	
/*	    file		    				    */
/* input:   HTTPNotifyConfig;   					    */
/* return:  LAN_OK - success					    */	
/* 	    LAN_ERROR -failure					    */
/*------------------------------------------------------------------*/

int LANReadHTTP_NOTIFY(HTTPNotifyConfig *info);

/*------------------------------------------------------------------*/
/* purpose: Write HTTP IP address notification to configuration     */ 	
/*	    file		    				    */
/* input:   HTTPNotifyConfig;   					    */
/* return:  LAN_OK - success					    */	
/* 	    LAN_ERROR -failure					    */
/*------------------------------------------------------------------*/

int LANWriteHTTP_NOTIFY(HTTPNotifyConfig *info);

#if 0
/*bulk cgi*/
struct log_bulk_ds
{
	SYSLOG_INFO syslog_server_ds;
	struct LogJabberConf jabberlog_ds;
	struct LogTypeConf logtype_ds;
};

int LOG_BULK_ReadConf(void *conf);
int LOG_BULK_SetConf(void* conf, char* item, char* value);
int LOG_BULK_CheckConf(void *conf);
int LOG_BULK_WriteConf(void *conf, void* conf_org);
int LOG_BULK_RunConf(void* conf, void* conf_org);

#endif

#endif
